/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "doca_sha_offload_lib.h"

#include <errno.h>
#include <getopt.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>

#include <openssl/opensslv.h>
#include <openssl/evp.h>
#include <openssl/crypto.h>
#include <openssl/async.h>

#include <doca_log.h>
#include <doca_version.h>
#include <doca_buf.h>
#include <doca_buf_inventory.h>
#include <doca_dev.h>
#include <doca_ctx.h>
#include <doca_mmap.h>
#include <doca_sha.h>
#include "doca_sha_job_id_generator.h"

DOCA_LOG_REGISTER(DOCA_SHA_OFFLOAD_LIB);

#define DOCA_SHA_MAX_PCI_ADDRESS_LEN 32
#define DOCA_SHA_MAX_WORKQ_NUM 1
#define DOCA_SHA_NB_ID 1024
#define DOCA_SHA_MAX_BUF_CNT ((DOCA_SHA_NB_ID) * 2)

/**
 * A helper structure for doca_buf inventory.
 */
struct doca_sha_registered_buf {
	const void *addr;
	size_t len;
};

/**
 * Data structure used to combining the doca_sha polling mode
 * and async_job library.
 */
enum doca_sha_id_status {
	DOCA_SHA_ID_STATUS_FREE = 0,
	DOCA_SHA_ID_STATUS_NOT_FREE,
	DOCA_SHA_ID_STATUS_REQ_SENT,
	DOCA_SHA_ID_STATUS_RESP_RECVD,
};

/**
 * Data structure used for maintaining doca_sha engine state.
 */
struct doca_sha_engine_state {
	struct doca_dev *dev;
	struct doca_mmap *mmap;
	struct doca_buf_inventory *buf_inv;
	struct doca_sha *sha_ctx;
	struct doca_ctx *ctx;
	struct doca_workq *workq[DOCA_SHA_MAX_WORKQ_NUM];
	uint8_t nb_workq;
	uint8_t workq_depth;
	bool already_init;
	uint32_t nb_doca_buf;
	uint32_t nb_inventory_buf;
	char pci_address[DOCA_SHA_MAX_PCI_ADDRESS_LEN + 1];
	struct doca_sha_registered_buf *global_reged_req_buf;
	struct doca_sha_registered_buf *global_reged_resp_buf;
	size_t global_reged_req_buf_cnt;
	size_t global_reged_resp_buf_cnt;
	unsigned char *resp_buf;
	struct doca_sha_partial_job *jobs;
	OSSL_ASYNC_FD *wfds;
	struct doca_sha_id_generator *id_alloc;
	enum doca_sha_id_status *id_status_array;
};

/**
 * Data structure used for doca_sha calculation.
 */
struct doca_sha_ctx {
	struct doca_sha_engine_state *state;
	struct doca_sha_partial_job *job;
	int sha_func;
};

static void
wait_cleanup_with_id(ASYNC_WAIT_CTX *ctx, const void *key, OSSL_ASYNC_FD r, void *vw)
{
	OSSL_ASYNC_FD *w = (OSSL_ASYNC_FD *)vw;

	close(r);
	close(*w);
}

static void
wait_cleanup_without_id(ASYNC_WAIT_CTX *ctx, const void *key, OSSL_ASYNC_FD r, void *vw)
{
	OSSL_ASYNC_FD *w = (OSSL_ASYNC_FD *)vw;

	close(r);
	close(*w);

	OPENSSL_free(w);
}

static void
doca_sha_pause_job(OSSL_ASYNC_FD *wfds, uint32_t id, int debug_level)
{
	ASYNC_JOB *job;
	ASYNC_WAIT_CTX *waitctx;
	OSSL_ASYNC_FD pipefds[2] = {0, 0};
	OSSL_ASYNC_FD *writefd;
	char buf = 'X';
	static const int unique;

	job = ASYNC_get_current_job();
	if (job != NULL) {
#ifdef PRINT_DEBUG_INFO
		if (debug_level > 1)
			DOCA_LOG_DBG("Executing within a job, so it is async_jobs mode");
#endif /* PRINT_DEBUG_INFO */
	} else {
#ifdef PRINT_DEBUG_INFO
		if (debug_level > 1)
			DOCA_LOG_DBG("Not executing within a job, so it is not aysnc_jobs mode");
#endif /* PRINT_DEBUG_INFO */
		return;
	}

	waitctx = ASYNC_get_wait_ctx(job);
	if (ASYNC_WAIT_CTX_get_fd(waitctx, &unique, &pipefds[0], (void **)&writefd))
		pipefds[1] = *writefd;
	else {
		if (id == DOCA_SHA_ID_GENERATOR_INVALID_ID)
			writefd = OPENSSL_malloc(sizeof(*writefd));
		else
			writefd = wfds + id - 1;
		if (writefd == NULL) {
			DOCA_LOG_ERR("Failed to malloc writefd");
			return;
		}
		if (pipe(pipefds) != 0) {
			DOCA_LOG_ERR("Failed to create pipe");
			if (id == DOCA_SHA_ID_GENERATOR_INVALID_ID)
				OPENSSL_free(writefd);
			return;
		}
		*writefd = pipefds[1];
		if (id == DOCA_SHA_ID_GENERATOR_INVALID_ID) {
			if (!ASYNC_WAIT_CTX_set_wait_fd(waitctx, &unique, pipefds[0], writefd,
				wait_cleanup_without_id)) {
				wait_cleanup_without_id(waitctx, &unique, pipefds[0], writefd);
				return;
			}
		} else {
			if (!ASYNC_WAIT_CTX_set_wait_fd(waitctx, &unique, pipefds[0], writefd, wait_cleanup_with_id)) {
				wait_cleanup_with_id(waitctx, &unique, pipefds[0], writefd);
				return;
			}
		}
	}
	/*
	 * In this async engine we use poll, not interrupt.
	 * So we need to signal that the job is ready to continue before the call to ASYNC_pause_job().
	 */
	if (write(pipefds[1], &buf, 1) < 0) {
		DOCA_LOG_ERR("write fd error");
		return;
	}

	/*
	 * Let the other async_jobs to run.
	 */
	if (ASYNC_pause_job() == 0) {
		DOCA_LOG_ERR("ASYNC_pause_job failure");
		return;
	}

	/* Clear the wake signal */
	if (read(pipefds[0], &buf, 1) < 0) {
		DOCA_LOG_ERR("read fd error");
		return;
	}
}

static inline char
convert_hex_to_lowcase_char(uint8_t hex_digit)
{
	char ret = hex_digit;

	if (hex_digit <= 9)
		ret += '0';
	else
		ret += 'a' - 10;
	return ret;
}

static inline struct doca_dev *
find_valid_local_device(const char *input_pci_addr, int debug_level)
{
	struct doca_devinfo **devinfo_list = NULL;
	struct doca_dev *ret_dev = NULL;
	int i = 0;
	int j = 0;
	uint32_t dev_count = 0;
	struct doca_pci_bdf pci_addr;
	char pci_addr_str[DOCA_SHA_MAX_PCI_ADDRESS_LEN + 1];

	if (doca_devinfo_list_create(&devinfo_list, &dev_count) != DOCA_SUCCESS) {
		DOCA_LOG_ERR("devinfo list create failure");
		return NULL;
	}
	for (i = 0; i < dev_count; i++) {
		if (!(devinfo_list[i]))
			goto err_find_dev;
		if (input_pci_addr != NULL) {
			if (doca_devinfo_get_pci_addr(devinfo_list[i], &pci_addr) != DOCA_SUCCESS)
				continue;

			memset(pci_addr_str, 0, sizeof(pci_addr_str));
			j = 0;
			pci_addr_str[j++] = convert_hex_to_lowcase_char((pci_addr.bus >> 4) & 0x0f);
			pci_addr_str[j++] = convert_hex_to_lowcase_char(pci_addr.bus & 0x0f);
			pci_addr_str[j++] = ':';
			pci_addr_str[j++] = convert_hex_to_lowcase_char((pci_addr.device >> 4) & 0x0f);
			pci_addr_str[j++] = convert_hex_to_lowcase_char(pci_addr.device & 0x0f);
			pci_addr_str[j++] = '.';
			pci_addr_str[j++] = convert_hex_to_lowcase_char(pci_addr.function & 0x0f);
#ifdef PRINT_DEBUG_INFO
			if (debug_level > 1)
				DOCA_LOG_DBG("The [%d]-th dev pci addr str = %s", i, pci_addr_str);
#endif /* PRINT_DEBUG_INFO */
			if (strncmp(pci_addr_str, input_pci_addr, DOCA_SHA_MAX_PCI_ADDRESS_LEN) == 0) {
				if (doca_sha_get_hardware_supported(devinfo_list[i]) != DOCA_SUCCESS)
					goto err_find_dev;
				if (doca_sha_job_get_supported(devinfo_list[i], DOCA_SHA_JOB_SHA1) != DOCA_SUCCESS)
					goto err_find_dev;
				if (doca_sha_job_get_supported(devinfo_list[i], DOCA_SHA_JOB_SHA256) != DOCA_SUCCESS)
					goto err_find_dev;
				if (doca_sha_job_get_supported(devinfo_list[i], DOCA_SHA_JOB_SHA512) != DOCA_SUCCESS)
					goto err_find_dev;
				if (doca_dev_open(devinfo_list[i], &ret_dev) == DOCA_SUCCESS) {
					doca_devinfo_list_destroy(devinfo_list);
					return ret_dev;
				}
			}
		} else {
			if (doca_sha_get_hardware_supported(devinfo_list[i]) != DOCA_SUCCESS)
				continue;
			if (doca_sha_job_get_supported(devinfo_list[i], DOCA_SHA_JOB_SHA1) != DOCA_SUCCESS)
				continue;
			if (doca_sha_job_get_supported(devinfo_list[i], DOCA_SHA_JOB_SHA256) != DOCA_SUCCESS)
				continue;
			if (doca_sha_job_get_supported(devinfo_list[i], DOCA_SHA_JOB_SHA512) != DOCA_SUCCESS)
				continue;
			if (doca_dev_open(devinfo_list[i], &ret_dev) == DOCA_SUCCESS) {
				doca_devinfo_list_destroy(devinfo_list);
				return ret_dev;
			}
		}
	}
err_find_dev:
	doca_devinfo_list_destroy(devinfo_list);
	return NULL;
}

static inline void
destroy_objects(struct doca_sha_engine_state *state)
{
	int i = 0;

	if (state == NULL)
		return;

#ifdef OPENSSL_BUG_FIXED
	if (state->sha_ctx != NULL) {
		doca_sha_destroy(state->sha_ctx);
		state->sha_ctx = NULL;
		state->ctx = NULL;
	}
#endif /* OPENSSL_BUG_FIXED */
	for (i = 0; i < state->nb_workq; i++) {
		if (state->workq[i] != NULL) {
			doca_workq_destroy(state->workq[i]);
			state->workq[i] = NULL;
		}
	}
	if (state->buf_inv != NULL) {
		doca_buf_inventory_destroy(state->buf_inv);
		state->buf_inv = NULL;
	}
#ifdef OPENSSL_BUG_FIXED
	if (state->dev != NULL) {
		doca_dev_close(state->dev);
		state->dev = NULL;
	}
#endif /* OPENSSL_BUG_FIXED */
	if (state->mmap != NULL) {
		doca_mmap_destroy(state->mmap);
		state->mmap = NULL;
	}
}

static inline bool
cleanup_objects(struct doca_sha_engine_state *state)
{
	doca_error_t ret;
	int i;

	ret = doca_buf_inventory_stop(state->buf_inv);
	if (ret != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Unable to stop inventory due to [%s]!",
			doca_get_error_string(ret));
		return false;
	}
	ret = doca_mmap_dev_rm(state->mmap, state->dev);
	if (ret != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Unable to de-register dev from mmap due to [%s]!",
			doca_get_error_string(ret));
		return false;
	}
	ret = doca_mmap_stop(state->mmap);
	if (ret != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Unable to stop mmap due to [%s]!",
			doca_get_error_string(ret));
		return false;
	}
	for (i = 0; i < state->nb_workq; i++) {
		ret = doca_ctx_workq_rm(state->ctx, state->workq[i]);
		if (ret != DOCA_SUCCESS) {
			DOCA_LOG_ERR("Unable to de-register workq from context due to [%s]!",
				doca_get_error_string(ret));
			return false;
		}
	}
	ret = doca_ctx_stop(state->ctx);
	if (ret != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Unable to stop context due to [%s]!",
			doca_get_error_string(ret));
		return false;
	}
	ret = doca_ctx_dev_rm(state->ctx, state->dev);
	if (ret != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Unable to de-register dev from context due to [%s]!",
			doca_get_error_string(ret));
		return false;
	}
	return true;
}

static inline bool
create_objects(struct doca_sha_engine_state *state, int debug_level)
{
	int i = 0;
	doca_error_t ret;

	if (state == NULL)
		return false;

	state->dev = find_valid_local_device(state->pci_address, debug_level);
	if (state->dev == NULL) {
		state->dev = find_valid_local_device(NULL, debug_level);
		if (state->dev == NULL) {
			DOCA_LOG_ERR("No suitable DOCA device found!");
			return false;
		}
	}

	/* The purpose of *2 is to make sure count both request buf and response buf */
	if (doca_buf_inventory_create(NULL,
		(state->nb_doca_buf * 2),
		DOCA_BUF_EXTENSION_NONE, &state->buf_inv) != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Unable to create DOCA buf inventory!");
		return false;
	}

	for (i = 0; i < state->nb_workq; i++) {
		if (doca_workq_create(state->workq_depth, &state->workq[i]) != DOCA_SUCCESS) {
			DOCA_LOG_ERR("Unable to create DOCA workq!");
			return false;
		}
	}

	if (doca_sha_create(&state->sha_ctx) != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Unable to create DOCA SHA context!");
		return false;
	}
	state->ctx = doca_sha_as_ctx(state->sha_ctx);

	ret = doca_ctx_dev_add(state->ctx, state->dev);
	if (ret != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Unable to register device with context due to [%s]!",
			doca_get_error_string(ret));
		return false;
	}
	ret = doca_ctx_start(state->ctx);
	if (ret != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Unable to start context due to [%s]!",
			doca_get_error_string(ret));
		return false;
	}
	for (i = 0; i < state->nb_workq; i++) {
		ret = doca_ctx_workq_add(state->ctx, state->workq[i]);
		if (ret != DOCA_SUCCESS) {
			DOCA_LOG_ERR("Unable to register workq with context due to [%s]!",
				doca_get_error_string(ret));
			return false;
		}
	}
	ret = doca_buf_inventory_start(state->buf_inv);
	if (ret != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Unable to start doca buf inventory due to [%s]!",
			doca_get_error_string(ret));
		return false;
	}
	return true;
}

/**
 * To do proper initialization to the above standard and special structs
 */
static inline bool
init_objects(struct doca_sha_engine_state *state, const uint32_t nb_input_buf)
{
	const uint32_t max_chunks = nb_input_buf;

	if (doca_mmap_create(NULL, &state->mmap) != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Unable to create DOCA mmap!");
		return false;
	}
	if (doca_mmap_set_max_num_chunks(state->mmap, max_chunks) != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Unable to set DOCA SHA mmap nb chunks!");
		return false;
	}
	if (doca_mmap_start(state->mmap) != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Unable to start DOCA SHA mmap!");
		return false;
	}
	if (doca_mmap_dev_add(state->mmap, state->dev) != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Unable to add device to mmap!");
		return false;
	}

	return true;
}

static inline bool
populate_mmap(struct doca_sha_engine_state *state, const unsigned char *addr, size_t len,
	      struct doca_sha_registered_buf *global_reged_buf, size_t *global_reged_buf_cnt, int debug_level)
{
	bool need_populate = true;
	size_t i;

#ifdef PRINT_DEBUG_INFO
	size_t j;

	if (debug_level > 1)
		DOCA_LOG_DBG("to populate: buf.addr = %p, buf.len = %ld, global_reged_buf_cnt = %ld",
			addr, len, *global_reged_buf_cnt);
#endif /* PRINT_DEBUG_INFO */

	for (i = 0; i < *global_reged_buf_cnt; i++) {
		if ((global_reged_buf[i].addr == (void *)addr) && (global_reged_buf[i].len == len)) {
			need_populate = false;
			break;
		}
	}
	if (need_populate) {
		if (doca_mmap_populate(state->mmap, (char *)(addr), len,
			sysconf(_SC_PAGESIZE), NULL, NULL) != DOCA_SUCCESS) {
			DOCA_LOG_ERR("Unable to populate DOCA SHA mmap with input buffer!");
			DOCA_LOG_ERR("error occurs to populate: buf.addr = %p, buf.len = %ld, global_reged_buf_cnt = %ld",
				addr, len, *global_reged_buf_cnt);
#ifdef PRINT_DEBUG_INFO
			if (debug_level > 1) {
				for (j = 0; j < *global_reged_buf_cnt; j++) {
					DOCA_LOG_ERR("[%ld]-th populated: buf.addr = %p, buf.len = %ld, global_reged_buf_cnt = %ld",
						j, global_reged_buf[j].addr, global_reged_buf[j].len, *global_reged_buf_cnt);
				}
			}
#endif /* PRINT_DEBUG_INFO */
			return false;
		}
		global_reged_buf[*global_reged_buf_cnt].addr = addr;
		global_reged_buf[*global_reged_buf_cnt].len = len;
		(*global_reged_buf_cnt)++;
	}
	return true;
}

int
doca_sha_reset_engine_state(struct doca_sha_engine_state *state, int debug_level)
{
	if (state == NULL) {
		DOCA_LOG_ERR("Error: doca-sha-engine-state ptr is null");
		return 1;
	}
	state->already_init = false;
	return 0;
}

int
doca_sha_engine_create(struct doca_sha_engine_state **out_state, int debug_level)
{
	struct doca_sha_engine_state *state = NULL;

	if (out_state == NULL) {
		DOCA_LOG_ERR("Error: out state is null");
		return 0;
	}
	state = calloc(1, sizeof(*state));
	if (state == NULL)
		return 1;
	*out_state = state;
	return 0;
}

int
doca_sha_engine_init(struct doca_sha_engine_state *state, const char *pci_addr, int debug_level)
{
	if (state == NULL) {
		DOCA_LOG_ERR("Error: doca-sha-engine-state ptr is null");
		return 0;
	}
	if (state->already_init == true) {
		DOCA_LOG_ERR("Error: doca-sha-engine already inited");
		return 0;
	}

	/*
	 * set default value
	 */
	state->nb_doca_buf = DOCA_SHA_MAX_BUF_CNT;
	memset(state->pci_address, 0, sizeof(state->pci_address));
	/*
	 * if user set the pci_addr, we use it as the engine's pci_addr
	 */
	if (pci_addr != NULL)
		memcpy(state->pci_address, pci_addr, 7);
	state->nb_workq = 1;
	state->workq_depth = 1;

	/* preparation-step-1: Create a doca-sha instance */
	if (!create_objects(state, debug_level))
		goto err_create_object;

	if (!init_objects(state, state->nb_doca_buf))
		goto err_init_object;

	state->resp_buf = calloc(DOCA_SHA512_BYTE_COUNT * DOCA_SHA_NB_ID, sizeof(uint8_t));
	if (state->resp_buf == NULL) {
		DOCA_LOG_ERR("Failed to allocate resp_buf raw data memory.");
		goto err_alloc_resp_buf;
	}

	state->jobs = calloc(DOCA_SHA_NB_ID, sizeof(struct doca_sha_partial_job));
	if (state->jobs == NULL) {
		DOCA_LOG_ERR("Failed to allocate jobs memory.");
		goto err_alloc_jobs;
	}

	state->wfds = OPENSSL_malloc(DOCA_SHA_NB_ID * sizeof(OSSL_ASYNC_FD));
	if (state->wfds == NULL) {
		DOCA_LOG_ERR("Failed to allocate wfds");
		goto err_alloc_wfds;
	}

	state->id_alloc = doca_sha_id_generator_create(DOCA_SHA_NB_ID);
	if (state->id_alloc == NULL) {
		DOCA_LOG_ERR("Failed to allocate id-alloc.");
		goto err_id_alloc_create;
	}
	state->id_status_array = calloc(DOCA_SHA_NB_ID, sizeof(enum doca_sha_id_status));
	if (state->id_status_array == NULL) {
		DOCA_LOG_ERR("Failed to allocate id-status-array.");
		goto err_id_status_malloc;
	}

	state->global_reged_req_buf_cnt = 0;
	state->global_reged_resp_buf_cnt = 0;
	state->global_reged_req_buf = calloc(DOCA_SHA_NB_ID, sizeof(struct doca_sha_registered_buf));
	if (state->global_reged_req_buf == NULL) {
		DOCA_LOG_ERR("Failed to allocate global_reged_req_buf.");
		goto err_global_reged_req_buf;
	}
	state->global_reged_resp_buf = calloc(DOCA_SHA_NB_ID, sizeof(struct doca_sha_registered_buf));
	if (state->global_reged_resp_buf == NULL) {
		DOCA_LOG_ERR("Failed to allocate global_reged_resp_buf.");
		goto err_global_reged_resp_buf;
	}

	if (!(populate_mmap(state, state->resp_buf, DOCA_SHA512_BYTE_COUNT * DOCA_SHA_NB_ID,
			    state->global_reged_resp_buf, &(state->global_reged_resp_buf_cnt), debug_level))) {
		DOCA_LOG_ERR("Failed to populate mmap.");
		goto err_populate_mmap;
	}

	state->already_init = true;

	return 1;
err_global_reged_resp_buf:
	free(state->global_reged_req_buf);
err_global_reged_req_buf:
	free(state->id_status_array);
err_id_status_malloc:
	doca_sha_id_generator_destroy(state->id_alloc);
err_id_alloc_create:
	free(state->jobs);
err_alloc_jobs:
	OPENSSL_free(state->wfds);
err_alloc_wfds:
err_populate_mmap:
	free(state->resp_buf);
err_alloc_resp_buf:
err_init_object:
	cleanup_objects(state);
err_create_object:
	destroy_objects(state);
	return 0;
}

int
doca_sha_engine_finish(struct doca_sha_engine_state *state, int debug_level)
{
	if (state == NULL) {
		DOCA_LOG_ERR("Error: state is null");
		return 0;
	}
	if (state->resp_buf != NULL)
		free(state->resp_buf);
	if (state->jobs != NULL)
		free(state->jobs);
	if (state->wfds != NULL)
		OPENSSL_free(state->wfds);
	doca_sha_id_generator_destroy(state->id_alloc);
	if (state->id_status_array != NULL)
		free(state->id_status_array);
	cleanup_objects(state);
	destroy_objects(state);
	state->already_init = false;
	return 1;
}

int
doca_sha_engine_destroy(struct doca_sha_engine_state *state, int debug_level)
{
	if (state == NULL) {
		DOCA_LOG_ERR("Error: state is null");
		return 0;
	}
	if (state->already_init == true) {
		destroy_objects(state);
		state->already_init = false;
	}
	free(state);
	return 1;
}

static inline uint8_t
doca_sha_get_response_byte_count(int sha_func)
{
	uint8_t resp_len = 0;

	switch (sha_func) {
	case DOCA_SHA_FUNC_TYPE_SHA1:
		resp_len = DOCA_SHA1_DIGEST_LEN;
		break;
	case DOCA_SHA_FUNC_TYPE_SHA256:
		resp_len = DOCA_SHA256_DIGEST_LEN;
		break;
	case DOCA_SHA_FUNC_TYPE_SHA512:
		resp_len = DOCA_SHA512_DIGEST_LEN;
		break;
	default:
		break;
	}
	return resp_len;
}

#ifdef PRINT_DEBUG_INFO
static int
report_results(const unsigned char *input, unsigned int input_len,
	       const struct doca_sha_partial_job *job)
{
	int i;
	int sha_func = DOCA_SHA_FUNC_TYPE_SHA1;
	void *data_addr = NULL;
	size_t data_len = 0;

	if (input != NULL) {
		printf("--- input data first 64bytes hex format is:");
		for (i = 0; (i < input_len) && (i < 64); i++)
			printf("%02x", input[i]);
		printf("\n");
	}
	if (job != NULL) {
		switch (job->sha_job.base.type) {
		case DOCA_SHA_JOB_SHA1:
		case DOCA_SHA_JOB_SHA1_PARTIAL:
			sha_func = DOCA_SHA_FUNC_TYPE_SHA1;
			break;
		case DOCA_SHA_JOB_SHA256:
		case DOCA_SHA_JOB_SHA256_PARTIAL:
			sha_func = DOCA_SHA_FUNC_TYPE_SHA256;
			break;
		case DOCA_SHA_JOB_SHA512:
		case DOCA_SHA_JOB_SHA512_PARTIAL:
			sha_func = DOCA_SHA_FUNC_TYPE_SHA512;
			break;
		default:
			break;
		}
		printf("--- job input data (only first 64bytes) hex format is:");
		if (doca_buf_get_data(job->sha_job.req_buf, &data_addr) != DOCA_SUCCESS) {
			DOCA_LOG_ERR("doca_buf_get_data() req_buf addr failure");
			return -EIO;
		}
		if (doca_buf_get_data_len(job->sha_job.req_buf, &data_len) != DOCA_SUCCESS) {
			DOCA_LOG_ERR("doca_buf_get_data_len() req_buf length failure");
			return -EIO;
		}
		for (i = 0; (i < data_len) && (i < 64); i++)
			printf("%02x", ((uint8_t *)data_addr)[i] & 0xff);
		printf("\n");
		printf("--- job received sha hex format is:\n");
		if (doca_buf_get_data(job->sha_job.resp_buf, &data_addr) != DOCA_SUCCESS) {
			DOCA_LOG_ERR("doca_buf_get_data() resp_buf addr failure");
			return -EIO;
		}
		if (doca_buf_get_data_len(job->sha_job.resp_buf, &data_len) != DOCA_SUCCESS) {
			DOCA_LOG_ERR("doca_buf_get_data_len() resp_buf length failure");
			return -EIO;
		}
		if (data_len != doca_sha_get_response_byte_count(sha_func))
			return -EIO;
		for (i = 0; i < data_len; i++)
			printf("%02x", ((uint8_t *)data_addr)[i] & 0xff);
		printf("\n");
	}
	return 0;
}
#endif /* PRINT_DEBUG_INFO */

static inline int
doca_sha_cal(struct doca_sha_ctx *ctx, unsigned char *input, unsigned int input_len, int debug_level)
{
	doca_error_t submit_ret;
	doca_error_t progress_ret;
	struct doca_event event = {0};
	uint16_t refcnt = 0;
	uint32_t id = DOCA_SHA_ID_GENERATOR_INVALID_ID;
	uint32_t recv_id = DOCA_SHA_ID_GENERATOR_INVALID_ID;

#ifdef PRINT_DEBUG_INFO
	if (debug_level > 1)
		DOCA_LOG_DBG("enter [%s], ctx = [%p]", __func__, ctx);
#endif /* PRINT_DEBUG_INFO */

	/*
	 * First try to get a id.
	 */
	while (1) {
		id = doca_sha_id_generator_get(ctx->state->id_alloc);
		if (id == DOCA_SHA_ID_GENERATOR_INVALID_ID)
			doca_sha_pause_job(ctx->state->wfds, id, debug_level);
		else {
			/*
			 * Because the allocated id starts from 1, so we need minus 1.
			 */
			ctx->state->id_status_array[id - 1] = DOCA_SHA_ID_STATUS_NOT_FREE;
			break;
		}
	}
	ctx->job = ctx->state->jobs + id - 1;
	if (!(populate_mmap(ctx->state, input, input_len,
		ctx->state->global_reged_req_buf, &(ctx->state->global_reged_req_buf_cnt), debug_level))) {
		DOCA_LOG_ERR("Failed to populate mmap .");
		goto err_populate_mmap;
	}
	/*
	 * create job.
	 * use user_data.u64 to transmit the id
	 */
	ctx->job->sha_job.base = (struct doca_job) {
		.type = DOCA_SHA_JOB_SHA1,
		.flags = DOCA_JOB_FLAGS_NONE,
		.ctx = ctx->state->ctx,
		.user_data.u64 = id,
	};
	if (ctx->sha_func == DOCA_SHA_FUNC_TYPE_SHA256)
		ctx->job->sha_job.base.type = DOCA_SHA_JOB_SHA256;
	else if (ctx->sha_func == DOCA_SHA_FUNC_TYPE_SHA512)
		ctx->job->sha_job.base.type = DOCA_SHA_JOB_SHA512;
	if (doca_buf_inventory_buf_by_addr(ctx->state->buf_inv, ctx->state->mmap,
		(char *)(input), input_len,
		&(ctx->job->sha_job.req_buf)) != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Unable to acquire DOCA machine accessible buffer representing req buffer!");
		goto err_doca_sha_cal;
	}
	if (doca_buf_set_data(ctx->job->sha_job.req_buf, input, input_len) != DOCA_SUCCESS) {
		DOCA_LOG_ERR("doca_buf_set_data() for req failure");
		goto err_doca_sha_cal;
	}
	if (doca_buf_inventory_buf_by_addr(ctx->state->buf_inv, ctx->state->mmap,
		(char *)(ctx->state->resp_buf + (id - 1) * DOCA_SHA512_BYTE_COUNT),
		DOCA_SHA512_BYTE_COUNT,
		&(ctx->job->sha_job.resp_buf)) != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Unable to acquire DOCA machine accessible buffer representing resp buffer!");
		goto err_doca_sha_cal;
	}
	ctx->job->session = NULL;
	ctx->job->sha_job.flags = DOCA_SHA_JOB_FLAGS_NONE;

	/* Submit the job to let SHA engine work */
	while ((submit_ret = doca_workq_submit(ctx->state->workq[0], &(ctx->job->sha_job.base))) == DOCA_ERROR_NO_MEMORY)
		doca_sha_pause_job(ctx->state->wfds, id, debug_level);
	if (submit_ret != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to submit SHA job due to [%s]",
			doca_get_error_string(submit_ret));
		goto err_doca_sha_cal;
	}

	ctx->state->id_status_array[id - 1] = DOCA_SHA_ID_STATUS_REQ_SENT;

	/*
	 * Wait for job completion event
	 * two case:
	 * 1. the received_id == id, it is the response of the above reqeust.
	 * 2. the recevied_id != id, it is the response of another request in other async_jobs.
	 */
	while (1) {
		progress_ret = doca_workq_progress_retrieve(ctx->state->workq[0], &event,
			DOCA_WORKQ_RETRIEVE_FLAGS_NONE);
		if (progress_ret == DOCA_ERROR_AGAIN) {
			if (ctx->state->id_status_array[id - 1] == DOCA_SHA_ID_STATUS_REQ_SENT)
				doca_sha_pause_job(ctx->state->wfds, id, debug_level);
			else if (ctx->state->id_status_array[id - 1] == DOCA_SHA_ID_STATUS_RESP_RECVD) {
				/*
				 * SHA job event received successfully
				 * by another async_job.
				 */
				if ((int)(event.type) != (int)DOCA_ACTION_NONE) {
					DOCA_LOG_ERR("received event wrong");
					goto err_doca_sha_cal;
				}
				break;
			}
		}
		/* Verify a successful event was retrieved. */
		else if (progress_ret != DOCA_SUCCESS) {
			DOCA_LOG_ERR("Failed to progress response due to [%s]",
				doca_get_error_string(progress_ret));
			goto err_doca_sha_cal;
		} else {
			/*
			 * progress_ret == DOCA_SUCCESS.
			 * two case:
			 * 1. the received_id == id, it is the response of the above reqeust.
			 * 2. the recevied_di != id, it is the response of another request in other async_jobs.
			 */
			recv_id = event.user_data.u64;
			ctx->state->id_status_array[recv_id - 1] = DOCA_SHA_ID_STATUS_RESP_RECVD;
			if (recv_id != id)
				doca_sha_pause_job(ctx->state->wfds, id, debug_level);
			else {
				/* SHA job event received successfully */
				break;
			}
		}
	}
	ctx->state->id_status_array[id - 1] = DOCA_SHA_ID_STATUS_FREE;
	doca_sha_id_generator_put(ctx->state->id_alloc, id);
#ifdef PRINT_DEBUG_INFO
	if (debug_level > 1) {
		/* finally, display the received sha result */
		report_results(input, input_len, ctx->job);
	}
#endif /* PRINT_DEBUG_INFO */
	doca_buf_get_refcount(ctx->job->sha_job.req_buf, &(refcnt));
	doca_buf_refcount_rm(ctx->job->sha_job.req_buf, &(refcnt));
	doca_buf_get_refcount(ctx->job->sha_job.resp_buf, &(refcnt));
	doca_buf_refcount_rm(ctx->job->sha_job.resp_buf, &(refcnt));
	return 1;
err_doca_sha_cal:
	doca_buf_get_refcount(ctx->job->sha_job.req_buf, &(refcnt));
	doca_buf_refcount_rm(ctx->job->sha_job.req_buf, &(refcnt));
	ctx->job->sha_job.req_buf = NULL;
	doca_buf_get_refcount(ctx->job->sha_job.resp_buf, &(refcnt));
	doca_buf_refcount_rm(ctx->job->sha_job.resp_buf, &(refcnt));
	ctx->job->sha_job.resp_buf = NULL;
err_populate_mmap:
	cleanup_objects(ctx->state);
	doca_sha_id_generator_put(ctx->state->id_alloc, id);
	return 0;
}

int
doca_sha_init_v1(struct doca_sha_ctx *ctx, enum doca_sha_func_type sha_func, struct doca_sha_engine_state *state, int debug_level)
{
#ifdef PRINT_DEBUG_INFO
	if (debug_level > 1)
		DOCA_LOG_DBG("enter [%s], ctx = [%p], state = [%p]", __func__, ctx, state);
#endif /* PRINT_DEBUG_INFO */
	if (ctx == NULL || state == NULL) {
		DOCA_LOG_ERR("Error: ctx is null");
		return 0;
	}

	if (ctx->state == NULL)
		ctx->state = state;
	ctx->sha_func = sha_func;

	return 1;
}

int
doca_sha_update_v1(struct doca_sha_ctx *ctx, unsigned char *input, size_t input_len, int debug_level)
{
	int ret = 1;

#ifdef PRINT_DEBUG_INFO
	if (debug_level > 1)
		DOCA_LOG_DBG("enter [%s], ctx = [%p]", __func__, ctx);
#endif /* PRINT_DEBUG_INFO */

	if (ctx == NULL || input == NULL) {
		DOCA_LOG_ERR("Error: ctx or input is null");
		return 0;
	}
	if (input_len == 0) {
		DOCA_LOG_ERR("input-len == 0 is not allowed for doca_sha hardware engine");
		return 0;
	}

	ret = doca_sha_cal(ctx, input, input_len, debug_level);
	if (ret == 0)
		DOCA_LOG_ERR("do sha update failure");

	return ret;
}

int
doca_sha_final_v1(unsigned char *digest, struct doca_sha_ctx *ctx, int debug_level)
{
	int ret = 1;
	void *data_addr = NULL;

#ifdef PRINT_DEBUG_INFO
	if (debug_level > 1)
		DOCA_LOG_DBG("enter [%s], ctx = [%p]", __func__, ctx);
#endif /* PRINT_DEBUG_INFO */

	if (ctx == NULL) {
		DOCA_LOG_ERR("Error: ctx is null");
		return 0;
	}
	if (digest == NULL) {
		DOCA_LOG_ERR("Error: digest is null");
		return 0;
	}
	if (doca_buf_get_data(ctx->job->sha_job.resp_buf, &data_addr) != DOCA_SUCCESS) {
		DOCA_LOG_ERR("doca_buf_get_data() resp_buf addr failure");
		return 0;
	}
	memcpy(digest, data_addr, doca_sha_get_response_byte_count(ctx->sha_func));

	return ret;
}

int
doca_sha_copy_v1(struct doca_sha_ctx *to, const struct doca_sha_ctx *from, int debug_level)
{
#ifdef PRINT_DEBUG_INFO
	if (debug_level > 1)
		DOCA_LOG_DBG("enter [%s], to = [%p], from = [%p]", __func__, to, from);
#endif /* PRINT_DEBUG_INFO */

	if ((to == NULL) || (from == NULL))
		return 0;

	memcpy(to, from, sizeof(struct doca_sha_ctx));
	return 1;
}

size_t get_size_of_doca_sha_ctx(void)
{
	return sizeof(struct doca_sha_ctx);
}
